"""
    Pipeline 处理 item，清洗、入库

    注意：默认报错是 logger.error 输出，需要详细信息请使用 logger.exception（输出会很多，但很细）
"""
import palp
from loguru import logger


class Pipeline(palp.Pipeline):
    def pipeline_in(self, spider, item) -> None:
        """
        入库之前的操作，一般是清洗

        :param spider:
        :param item:
        :return:
        """

    def pipeline_save(self, spider, item) -> None:
        """
        入库

        :param spider:
        :param item: 启用 item_buffer 将会是 List[item] 反之为 item
        :return:
        """
        logger.info(item)

    def pipeline_error(self, spider, item, exception: Exception) -> None:
        """
        入库出错时的操作

        :param spider:
        :param item: 启用 item buffer 时是 List[item]
        :param exception: 错误的详细信息
        :return:
        """
        logger.error(exception)

    def pipeline_failed(self, spider, item) -> None:
        """
        超过最大重试次数时的操作

        :param spider:
        :param item: 启用 item buffer 时是 List[item]
        :return:
        """
        logger.warning(f"失败的 item：{item}")

    def pipeline_close(self, spider) -> None:
        """
        spider 结束时的操作，虽然是多线程，但是只会执行一次

        :param spider:
        :return:
        """



"""
    通过线程执行函数，简化创建线程
"""
import threading
from typing import Callable


class RunByThreadDecorator(threading.Thread):
    """
        通过线程执行函数，简化创建线程
    """

    def __init__(self, callback: Callable = None, **kwargs):
        super().__init__(**kwargs)
        self.func = None
        self.args = None
        self.kwargs = None
        self.callback = callback

    def run(self) -> None:
        res = self.func(*self.args, *self.kwargs)
        if self.callback:
            self.callback(res)

    def __call__(self, func: Callable):
        def _wrapper(*args, **kwargs):
            spider = args[0]
            if hasattr(spider, 'distribute_thread_list'):
                spider.distribute_thread_list.append(self)  # 将线程存入列表

            self.func = func
            self.args = args
            self.kwargs = kwargs
            self.start()

        return _wrapper



from palp.filter.filter_set import SetFilter
from palp.filter.filter_bloom import BloomFilter
from palp.filter.filter_redis_set import RedisSetFilter
from palp.filter.filter_redis_bloom import RedisBloomFilter



"""
    本地内存去重

    print(MemoryFilter.is_repeat(request=RequestGet(url='https://www.baidu.com')))
    print(MemoryFilter.is_repeat(request=RequestGet(url='https://www.baidu.com')))
"""
from palp import settings
from palp.network.request import Request
from palp.filter.filter import FilterBase, FilterLock


class SetFilter(FilterBase):
    def __init__(self):
        self.memory_filter_request = set()
        self.memory_filter_item = set()

    def is_repeat(self, obj, **kwargs):
        """
        获取对应的指纹，通过 python 的 set 去重

        :param obj:
        :param kwargs:
        :return:
        """
        fingerprint = self.fingerprint(obj=obj)

        if isinstance(obj, Request):
            memory_filter = self.memory_filter_request
        else:
            memory_filter = self.memory_filter_item

        if settings.STRICT_FILTER:
            with FilterLock():
                return self.judge(memory_filter, fingerprint)
        else:
            return self.judge(memory_filter, fingerprint)

    def judge(self, f, fingerprint):
        """
        进行判断

        :param f: 判断条件或方法之类
        :param fingerprint: 指纹
        :return:
        """
        if fingerprint in f:
            return True

        f.add(fingerprint)

        return False



"""
    爬虫执行结果汇总记录装饰器

    本身记录的过程是在中间件 RequestRecordMiddleware 中记录，并更新到 spider_record 变量上
    注意：
        这里侧重的爬虫执行结束之后的结果汇总

    爬虫执行完毕后这要做的事
        本地爬虫：无
        分布式爬虫：发送到 redis 最终多端合并
"""
import json
import datetime
from palp import settings


class SpiderRecordDecorator:
    """
        爬虫执行结果汇总记录装饰器
    """

    @staticmethod
    def record(spider):
        """
            进行记录
        """
        from palp.conn import redis_conn

        # 非分布式不上传 redis
        if settings.SPIDER_TYPE == 1:
            return

        # 发送每个 spider 的爬取情况
        redis_conn.hset(settings.REDIS_KEY_RECORD, spider.spider_uuid, json.dumps({
            'request_all': spider.spider_record['all'],
            'request_failed': spider.spider_record['failed'],
            'request_succeed': spider.spider_record['succeed'],
            'stop_time': str(datetime.datetime.now())
        }, ensure_ascii=False))

        # 删除当前 spider 心跳字段
        redis_conn.hdel(settings.REDIS_KEY_HEARTBEAT, spider.spider_uuid)

        # 如果是最后一个心跳，则进行统计记录
        if not redis_conn.exists(settings.REDIS_KEY_HEARTBEAT):
            request_all = 0
            request_failed = 0
            request_succeed = 0

            # 统计
            for spider_uuid, spider_detail in redis_conn.hgetall(settings.REDIS_KEY_RECORD).items():
                spider_detail = json.loads(spider_detail.decode())

                request_all += spider_detail['request_all']
                request_failed += spider_detail['request_failed']
                request_succeed += spider_detail['request_succeed']

            # 发送最后的统计结果
            redis_conn.set(settings.REDIS_KEY_STOP, json.dumps({
                'request_all': request_all,
                'request_failed': request_failed,
                'request_succeed': request_succeed,
                'stop_time': str(datetime.datetime.now())
            }, ensure_ascii=False))

    def __call__(self, func):
        def _wrapper(*args, **kwargs):
            # 先执行函数
            try:
                func(*args, **kwargs)
            finally:
                # 获取 spider
                spider = args[0]

                # 统计数据
                self.record(spider=spider)

        return _wrapper



"""
    周期爬虫记录请求中间件

    注意：
        双继承的中间件，需要导入 2 次
        双继承的中间件，文件名、类名不含有 middleware_spider、middleware_request
"""
import datetime
from palp import settings
from palp.spider.spider_cycle import CycleSpider
from palp.middleware.middleware_spider import SpiderMiddleware
from palp.middleware.middleware_request import RequestMiddleware


class CycleSpiderRecordRequestMiddleware(RequestMiddleware, SpiderMiddleware):
    """
        周期爬虫记录请求中间件
    """

    def spider_start(self, spider) -> None:
        self.start_time = datetime.datetime.now()

    def request_failed(self, spider, request):
        """
        失败请求记录

        :param spider:
        :param request:
        :return:
        """
        if issubclass(spider, CycleSpider) and hasattr(request, 'task'):
            spider.set_task_state_failed(task_id=request.task['id'])

    def request_close(self, spider, request, response):
        """
        成功请求记录

        :param spider:
        :param request:
        :return:
        """
        if issubclass(spider, CycleSpider) and hasattr(request, 'task'):
            spider.set_task_state_done(task_id=request.task['id'])

    def spider_close(self, spider) -> None:
        """
        写入记录表

        :param spider:
        :return:
        """
        from palp.conn import redis_conn

        if settings.SPIDER_TYPE == 1:
            pass
        else:
            pass



from palp.middleware.middleware_spider_recycle import SpiderRecycleMiddleware
from palp.middleware.middleware_request_recycle import RequestRecycleMiddleware
from palp.middleware.middleware_request_check import RequestCheckMiddleware
from palp.middleware.middleware_request_record import RequestRecordMiddleware
from palp.middleware.middleware_cycle_spider_record import CycleSpiderRecordRequestMiddleware
from palp.middleware.middleware_spider import SpiderMiddleware
from palp.middleware.middleware_request import RequestMiddleware
from palp.middleware.middleware_request_filter import RedisSetFilter, RedisBloomFilter, RedisSetFilterMiddleware, \
    RedisBloomFilterMiddleware



"""
    pipeline
"""
from typing import Union, List
from palp.item.item import Item
from palp.spider.spider import SpiderBase


class PipelineBase:
    """
        pipeline 基类
    """

    def pipeline_in(self, spider: SpiderBase, item: Item) -> None:
        """
        入库之前的操作

        :param spider:
        :param item:
        :return:
        """

    def pipeline_save(self, spider: SpiderBase, item: Union[Item, List[Item]]) -> None:
        """
        入库

        :param spider:
        :param item: 启用 item_buffer 将会是 List[item]
        :return:
        """

    def pipeline_error(self, spider: SpiderBase, item: Union[Item, List[Item]], exception: Exception) -> None:
        """
        入库出错时的操作

        :param spider:
        :param item: 启用 item buffer 时是 List[item]
        :param exception: 错误
        :return:
        """

    def pipeline_failed(self, spider: SpiderBase, item: Union[Item, List[Item]]) -> None:
        """
        超过最大重试次数时的操作

        :param spider:
        :param item: 启用 item buffer 时是 List[item]
        :return:
        """

    def pipeline_close(self, spider: SpiderBase) -> None:
        """
        spider 结束时的操作，虽然是多线程，但是只会执行一次

        :param spider:
        :return:
        """


class Pipeline(PipelineBase):
    """
        用于外部引用：pipeline
    """

    def pipeline_in(self, spider, item) -> None:
        """
        入库之前的操作

        :param spider:
        :param item:
        :return:
        """

    def pipeline_save(self, spider, item) -> None:
        """
        入库

        :param spider:
        :param item: 启用 item_buffer 将会是 List[item] 反之为 item
        :return:
        """

    def pipeline_error(self, spider, item, exception: Exception) -> None:
        """
        入库出错时的操作

        :param spider:
        :param item: 启用 item buffer 时是 List[item]
        :param exception: 错误
        :return:
        """

    def pipeline_failed(self, spider, item) -> None:
        """
        超过最大重试次数时的操作

        :param spider:
        :param item: 启用 item buffer 时是 List[item]
        :return:
        """

    def pipeline_close(self, spider) -> None:
        """
        spider 结束时的操作，虽然是多线程，但是只会执行一次

        :param spider:
        :return:
        """
